

/*************************************************************************/
/*                                                                       */
/*  Copyright (c) 1994 Stanford University                               */
/*                                                                       */
/*  All rights reserved.                                                 */
/*                                                                       */
/*  Permission is given to use, copy, and modify this software for any   */
/*  non-commercial purpose as long as this copyright notice is not       */
/*  removed.  All other uses, including redistribution in whole or in    */
/*  part, are forbidden without prior written permission.                */
/*                                                                       */
/*  This software is provided with absolutely no warranty and no         */
/*  support.                                                             */
/*                                                                       */
/*************************************************************************/

/*************************************************************************/
/*                                                                       */
/*  Parallel dense blocked LU factorization (no pivoting)                */
/*                                                                       */
/*  This version contains one dimensional arrays in which the matrix     */
/*  to be factored is stored.                                            */
/*                                                                       */
/*  Command line options:                                                */
/*                                                                       */
/*  -nN : Decompose NxN matrix.                                          */
/*  -pP : P = number of processors.                                      */
/*  -bB : Use a block size of B. BxB elements should fit in cache for    */
/*        good performance. Small block sizes (B=8, B=16) work well.     */
/*  -s  : Print individual processor timing statistics.                  */
/*  -t  : Test output.                                                   */
/*  -o  : Print out matrix values.                                       */
/*  -h  : Print out command line options.                                */
/*                                                                       */
/*  Note: This version works under both the FORK and SPROC models        */
/*                                                                       */
/*************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <rsim_apps.h>
TreeBar final_sync;

#define LU_FLAG 1

#define MAXRAND                         32767.0
#define DEFAULT_N                         256
#define DEFAULT_P                           1
#define DEFAULT_B                          16
#define min(a,b) ((a) < (b) ? (a) : (b))

#define PAGE_SIZE 4096
#define LINE_SIZE 64	
#define DOUBLES_PER_LINE 8	
#define NUMBER_OF_PROCESSORS P 

#ifdef LU_FLAG
#define flagnum(i,j) flag_array[(i)*nblocks+(j)]
#endif 

TreeBar final_sync;

struct GlobalMemory {
  double *t_in_fac;
  double *t_in_solve;
  double *t_in_mod;
  double *t_in_bar;
  double *completion;
  unsigned int starttime;
  unsigned int rf;
  unsigned int rs;
  unsigned int done;
  int id;
  TreeBar start;
  volatile int (idlock);
} *Global;

struct LocalCopies {
  double t_in_fac;
  double t_in_solve;
  double t_in_mod;
  double t_in_bar;
};

#ifdef LU_FLAG
typedef struct {
  int dummy00a[32];
  volatile int flag;
  int dummy00b[32];
} flag_struct;
#endif

int n = DEFAULT_N;          /* The size of the matrix */
int P = DEFAULT_P;          /* Number of processors */
int block_size = DEFAULT_B; /* Block dimension */
int nblocks;                /* Number of blocks in each dimension */
int num_rows;               /* Number of processors per row of processor grid */
int num_cols;               /* Number of processors per col of processor grid */
double *a;                  /* a = lu; l and u both placed back in a */
double *rhs;
int *proc_bytes;            /* Bytes to malloc per processor to hold blocks
                               of A*/
int test_result = 0;        /* Test result of factorization? */
int doprint = 0;            /* Print out matrix values? */
int dostats = 0;            /* Print out individual processor statistics? */
#ifdef LU_FLAG
int Istack, Jstack, Kstack;
#endif
int phase = 0;
void SlaveStart();
void OneSolve(int, int, double *, int, int);
void lu0(double *,int, int);
void bdiv(double *, double *, int, int, int, int);
void bmodd(double *, double*, int, int, int, int);
void bmod(double *, double *, double *, int, int, int, int);
void daxpy(double *, double *, int, double);
void daxpy2(double *, double *, int, double, int);
void daxpy2_pf(double *, double *,  int, double, int);
int BlockOwner(int, int);
void lu(int, int, int, struct LocalCopies *, int);
void InitA(double *);
double TouchA(int, int);
void PrintA();
void CheckResult(int, double *, double *);
void printerr(char *);

#ifdef LU_FLAG
flag_struct *flag_array;
#endif


main(argc, argv)

int argc;
char *argv[];

{
  int i, j;
  int ch;
  extern char *optarg;
  int MyNum=0;
  double mint, maxt, avgt;
  double min_fac, min_solve, min_mod, min_bar;
  double max_fac, max_solve, max_mod, max_bar;
  double avg_fac, avg_solve, avg_mod, avg_bar;
  int proc_num;
  unsigned int start;
  int a_pages, rhs_pages, doubles_in_page;
  int k;
  int flag_pages;


  {long time(long *); (start) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;

  while ((ch = getopt(argc, argv, "n:p:b:cstoh")) != -1) {
    switch(ch) {
    case 'n': n = atoi(optarg); break;
    case 'p': P = atoi(optarg); break;
    case 'b': block_size = atoi(optarg); break;
    case 's': dostats = 1; break;
    case 't': test_result = !test_result; break;
    case 'o': doprint = !doprint; break;
    case 'h': printf("Usage: LU <options>\n\n");
              printf("options:\n");
              printf("  -nN : Decompose NxN matrix.\n");
              printf("  -pP : P = number of processors.\n");
              printf("  -bB : Use a block size of B. BxB elements should fit in cache for \n");
              printf("        good performance. Small block sizes (B=8, B=16) work well.\n");
              printf("  -c  : Copy non-locally allocated blocks to local memory before use.\n");
              printf("  -s  : Print individual processor timing statistics.\n");
              printf("  -t  : Test output.\n");
              printf("  -o  : Print out matrix values.\n");
              printf("  -h  : Print out command line options.\n\n");
              printf("Default: LU -n%1d -p%1d -b%1d\n",
                     DEFAULT_N,DEFAULT_P,DEFAULT_B);
              exit(0);
              break;
    }
  }

  {{TreeBarInit(&(final_sync), NUMBER_OF_PROCESSORS);};}

  printf("\n");
  printf("Blocked Dense LU Factorization\n");
  printf("     %d by %d Matrix\n",n,n);
  printf("     %d Processors\n",P);
  printf("     %d by %d Element Blocks\n",block_size,block_size);
  printf("\n");
  printf("\n");

  num_rows = (int) sqrt((double) P);
  for (;;) {
    num_cols = P/num_rows;
    if (num_rows*num_cols == P)
      break;
    num_rows--;
  }
  nblocks = n/block_size;
  if (block_size * nblocks != n) {
    nblocks++;
  }

#ifdef LU_FLAG
  flag_pages = (nblocks*nblocks*sizeof(flag_struct))/PAGE_SIZE;
  if (flag_pages * PAGE_SIZE != nblocks*nblocks*sizeof(flag_struct))
    flag_pages++;
  flag_array = (flag_struct *)shmalloc(flag_pages*PAGE_SIZE + LINE_SIZE);;;
  flag_array = (flag_struct *)(((int)flag_array) + LINE_SIZE - (((unsigned)flag_array) % LINE_SIZE));

  for (k = 0; k < flag_pages; k++) {
    AssociateAddrNode((void *)(((unsigned)(flag_array))+k*PAGE_SIZE),
		      (void *)(((unsigned)(flag_array))+(k+1)*PAGE_SIZE),
                      k% P, "flag_node");
  }		      

  sys_bzero(flag_array, flag_pages*PAGE_SIZE);
  
#endif

  doubles_in_page = PAGE_SIZE/sizeof(double);
  a_pages = (n*n*sizeof(double))/PAGE_SIZE;
  if ( a_pages*PAGE_SIZE != n*n*sizeof(double) )
    a_pages++;

  a = (double *) shmalloc(n*n*sizeof(double));;
  /* a = (double *)(((int)a) + LINE_SIZE - (((unsigned)a) % LINE_SIZE)); */
  
  for (k = 0; k < a_pages; k++) {
    AssociateAddrNode(&a[k*doubles_in_page],
		      &a[(k+1)*doubles_in_page],
		      k % P, "a_node");
  }		
      
  rhs_pages = (n*sizeof(double))/PAGE_SIZE;
  if ( rhs_pages*PAGE_SIZE != n*sizeof(double) )
    rhs_pages++;
  rhs = (double *) shmalloc(n*sizeof(double));;
  /* rhs = (double *)(((int)rhs) + LINE_SIZE - (((unsigned)rhs) % LINE_SIZE)); */
  
  for (k = 0; k < rhs_pages; k++) {
    AssociateAddrNode(&rhs[k*doubles_in_page],
		      &rhs[(k+1)*(doubles_in_page)],
		      k % P, "rhs_node");
  }		      

  Global = (struct GlobalMemory *) shmalloc(sizeof(struct GlobalMemory));;
  AssociateAddrNode(Global,Global+1,
		    0, "g0_node");

  Global->t_in_fac = (double *) shmalloc(P*sizeof(double));;
  Global->t_in_mod = (double *) shmalloc(P*sizeof(double));;
  Global->t_in_solve = (double *) shmalloc(P*sizeof(double));;
  Global->t_in_bar = (double *) shmalloc(P*sizeof(double));;
  Global->completion = (double *) shmalloc(P*sizeof(double));;

  i = 0;
  AssociateAddrNode(Global->t_in_fac,Global->t_in_fac + P,
		    i++%P, "g1_node");
  AssociateAddrNode(Global->t_in_mod,Global->t_in_mod + P,
		    i++%P, "g2_node");
  AssociateAddrNode(Global->t_in_solve,Global->t_in_solve + P,
		    i++%P, "g3_node");
  AssociateAddrNode(Global->t_in_bar,Global->t_in_bar + P,
		    i++%P, "g4_node");
  AssociateAddrNode(Global->completion,Global->completion + P,
		    i++%P, "g5_node");
  
  if (Global == NULL) {
    printerr("Could not malloc memory for Global\n");
    exit(-1);
  } else if (Global->t_in_fac == NULL) {
    printerr("Could not malloc memory for Global->t_in_fac\n");
    exit(-1);
  } else if (Global->t_in_mod == NULL) {
    printerr("Could not malloc memory for Global->t_in_mod\n");
    exit(-1);
  } else if (Global->t_in_solve == NULL) {
    printerr("Could not malloc memory for Global->t_in_solve\n");
    exit(-1);
  } else if (Global->t_in_bar == NULL) {
    printerr("Could not malloc memory for Global->t_in_bar\n");
    exit(-1);
  } else if (Global->completion == NULL) {
    printerr("Could not malloc memory for Global->completion\n");
    exit(-1);
  }

/* POSSIBLE ENHANCEMENT:  Here is where one might distribute the a
   matrix data across physically distributed memories in a 
   round-robin fashion as desired. */

  {TreeBarInit(&(Global->start), NUMBER_OF_PROCESSORS);};
  {FREELOCK(&Global->idlock);};
  Global->id = 0;

  InitA(rhs);
  if (doprint) {
    printf("Matrix before decomposition:\n");
    PrintA();
  }

  MEMSYS_ON
    
  for (i=1; i<P; i++) {
    { if(fork() == 0) {(SlaveStart)(); TREEBAR(&final_sync, getpid()); exit(0);}}
  }


  SlaveStart(MyNum);

  {TREEBAR(&final_sync, getpid());}

  if (doprint) {
    printf("\nMatrix after decomposition:\n");
    PrintA();
  }

  if (dostats) {
    maxt = avgt = mint = Global->completion[0];
    for (i=1; i<P; i++) {
      if (Global->completion[i] > maxt) {
        maxt = Global->completion[i];
      }
      if (Global->completion[i] < mint) {
        mint = Global->completion[i];
      }
      avgt += Global->completion[i];
    }
    avgt = avgt / P;

    min_fac = max_fac = avg_fac = Global->t_in_fac[0];
    min_solve = max_solve = avg_solve = Global->t_in_solve[0];
    min_mod = max_mod = avg_mod = Global->t_in_mod[0];
    min_bar = max_bar = avg_bar = Global->t_in_bar[0];

    for (i=1; i<P; i++) {
      if (Global->t_in_fac[i] > max_fac) {
        max_fac = Global->t_in_fac[i];
      }
      if (Global->t_in_fac[i] < min_fac) {
        min_fac = Global->t_in_fac[i];
      }
      if (Global->t_in_solve[i] > max_solve) {
        max_solve = Global->t_in_solve[i];
      }
      if (Global->t_in_solve[i] < min_solve) {
        min_solve = Global->t_in_solve[i];
      }
      if (Global->t_in_mod[i] > max_mod) {
        max_mod = Global->t_in_mod[i];
      }
      if (Global->t_in_mod[i] < min_mod) {
        min_mod = Global->t_in_mod[i];
      }
      if (Global->t_in_bar[i] > max_bar) {
        max_bar = Global->t_in_bar[i];
      }
      if (Global->t_in_bar[i] < min_bar) {
        min_bar = Global->t_in_bar[i];
      }
      avg_fac += Global->t_in_fac[i];
      avg_solve += Global->t_in_solve[i];
      avg_mod += Global->t_in_mod[i];
      avg_bar += Global->t_in_bar[i];
    }
    avg_fac = avg_fac/P;
    avg_solve = avg_solve/P;
    avg_mod = avg_mod/P;
    avg_bar = avg_bar/P;
  }
  printf("                            PROCESS STATISTICS\n");
  printf("              Total      Diagonal     Perimeter      Interior       Barrier\n");
  printf(" Proc         Time         Time         Time           Time          Time\n");
  printf("    0    %10.0f    %10.0f    %10.0f    %10.0f    %10.0f\n",
          Global->completion[0],Global->t_in_fac[0],
          Global->t_in_solve[0],Global->t_in_mod[0],
          Global->t_in_bar[0]);
  if (dostats) {
    for (i=1; i<P; i++) {
      printf("  %3d    %10.0f    %10.0f    %10.0f    %10.0f    %10.0f\n",
              i,Global->completion[i],Global->t_in_fac[i],
              Global->t_in_solve[i],Global->t_in_mod[i],
              Global->t_in_bar[i]);
    }
    printf("  Avg    %10.0f    %10.0f    %10.0f    %10.0f    %10.0f\n",
           avgt,avg_fac,avg_solve,avg_mod,avg_bar);
    printf("  Min    %10.0f    %10.0f    %10.0f    %10.0f    %10.0f\n",
           mint,min_fac,min_solve,min_mod,min_bar);
    printf("  Max    %10.0f    %10.0f    %10.0f    %10.0f    %10.0f\n",
           maxt,max_fac,max_solve,max_mod,max_bar);
  }
  printf("\n");
  Global->starttime = start;
  printf("                            TIMING INFORMATION\n");
  printf("Start time                        : %16d\n",
          Global->starttime);
  printf("Initialization finish time        : %16d\n",
          Global->rs);
  printf("Overall finish time               : %16d\n",
          Global->rf);
  printf("Total time with initialization    : %16d\n",
          Global->rf-Global->starttime);
  printf("Total time without initialization : %16d\n",
          Global->rf-Global->rs);
  printf("\n");

  if (test_result) {
    printf("                             TESTING RESULTS\n");
    CheckResult(n, a, rhs);
  }

  {exit(0);};
}

void SlaveStart()

{
  int i;
  int j;
  int cluster;
  int max_block;
  int MyNum;
  /*
  {GETLOCK(&Global->idlock);}
    MyNum = Global->id;
    Global->id ++;
  {FREELOCK(&Global->idlock);}
  */

  MyNum = getpid(); 

#ifndef PHASE
  if(MyNum == 0){
    StatReportAll();
    StatClearAll();
  }
  newphase(++phase);
#endif


/* POSSIBLE ENHANCEMENT:  Here is where one might pin processes to
   processors to avoid migration */

  OneSolve(n, block_size, a, MyNum, dostats);
}


void OneSolve(n, block_size, a, MyNum, dostats)

double *a;
int n;
int block_size;
int MyNum;
int dostats;

{
  unsigned int i;
  unsigned int myrs;
  unsigned int myrf;
  unsigned int mydone;
  struct LocalCopies *lc;

  lc = (struct LocalCopies *) malloc(sizeof(struct LocalCopies));
  if (lc == NULL) {
    fprintf(stderr,"Proc %d could not malloc memory for lc\n",MyNum);
    exit(-1);
  }
  lc->t_in_fac = 0.0;
  lc->t_in_solve = 0.0;
  lc->t_in_mod = 0.0;
  lc->t_in_bar = 0.0;

  /* barrier to ensure all initialization is done */
  {TREEBAR(&(Global->start), MyNum);};

  /* to remove cold-start misses, all processors begin by touching a[] */
  TouchA(block_size, MyNum);

  {TREEBAR(&(Global->start), MyNum);};

/* POSSIBLE ENHANCEMENT:  Here is where one might reset the
   statistics that one is measuring about the parallel execution */

#ifndef PHASE
  endphase(phase);
  if(MyNum == 0){
    StatReportAll();
    StatClearAll();
  }
  newphase(++phase);
#endif  

  if ((MyNum == 0) || (dostats)) {
    {long time(long *); (myrs) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;
  }

  lu(n, block_size, MyNum, lc, dostats);

  if ((MyNum == 0) || (dostats)) {
    {long time(long *); (mydone) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;
  }

  {TREEBAR(&(Global->start), MyNum);};
#ifndef PHASE
  endphase(phase);
  if(MyNum == 0){
    StatReportAll();
    StatClearAll();
  }
#endif

  if ((MyNum == 0) || (dostats)) {
    {long time(long *); (myrf) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;
    Global->t_in_fac[MyNum] = lc->t_in_fac;
    Global->t_in_solve[MyNum] = lc->t_in_solve;
    Global->t_in_mod[MyNum] = lc->t_in_mod;
    Global->t_in_bar[MyNum] = lc->t_in_bar;
    Global->completion[MyNum] = mydone-myrs;
  }
  if (MyNum == 0) {
    Global->rs = myrs;
    Global->done = mydone;
    Global->rf = myrf;
  }
}


void lu0(a, n, stride)

double *a;
int n;
int stride;

{
  int i;
  int j; 
  int k;
  int length;
  double alpha;

  /* modify subsequent columns */
  
  for (k=0; k<n; k++) {
    for (j=k+1; j<n; j++) {
      a[k+j*stride] /= a[k+k*stride]; 
      alpha = -a[k+j*stride];
      length = n-k-1;
      daxpy(&a[k+1+j*stride], &a[k+1+k*stride], n-k-1, alpha); 
    }
  }
  
}


void bdiv(a, diag, stride_a, stride_diag, dimi, dimk)

double *a;
double *diag;
int stride_a;
int stride_diag;
int dimi;
int dimk;

{
  int i;
  int j;
  int k;
  double alpha;

#ifdef LU_FLAG      
  START_USR1;	
  WAITFLAG(&flagnum(Kstack,Kstack).flag,1);
  END_USR1;
#endif

  for (k=0; k<dimk; k++) {
    for (j=k+1; j<dimk; j++) {
      alpha = -diag[k+j*stride_diag];
      daxpy(&a[j*stride_a], &a[k*stride_a], dimi, alpha);
    }
  }
  
#ifdef LU_FLAG
  WRITEFLAG(&flagnum(Istack,Kstack).flag, 1);
#endif

}


void bmodd(a, c, dimi, dimj, stride_a, stride_c)

double *a;
double *c;
int dimi;
int dimj;
int stride_a;
int stride_c;

{
  int i;
  int j;
  int k;
  int length;
  double alpha;

#ifdef LU_FLAG      
  START_USR2;
  WAITFLAG(&flagnum(Kstack,Kstack).flag,1);
  END_USR2;
#endif

  for (k=0; k<dimi; k++)
    for (j=0; j<dimj; j++) {
      c[k+j*stride_c] /= a[k+k*stride_a];
      alpha = -c[k+j*stride_c];
      length = dimi - k - 1;
      daxpy(&c[k+1+j*stride_c], &a[k+1+k*stride_a], dimi-k-1, alpha);
    }

#ifdef LU_FLAG
  WRITEFLAG(&flagnum(Kstack,Jstack).flag,1);
#endif    

}


void bmod(a, b, c, dimi, dimj, dimk, stride)

double *a;
double *b;
double *c;
int dimi;
int dimj;
int dimk;
int stride;

{
  int i;
  int j;
  int k;
  double alpha;

#ifdef LU_FLAG
  START_USR3;
  WAITFLAG(&flagnum(Istack,Kstack).flag, 1);
  END_USR3;
  
  START_USR4;
  WAITFLAG(&flagnum(Kstack,Jstack).flag,1);
  END_USR4;
#endif

#ifndef CLUSTERED_MISS
  for (k=0; k<dimk; k++) {
    for (j=0; j<dimj; j++) {
      alpha = -b[k+j*stride];
      daxpy(&c[j*stride], &a[k*stride], dimi, alpha);
    }
  }
#else /* CLUSTERED MISS */
  for (i=0;i<dimi;i++) {
    for (k=0;k<dimi;k++) {
      alpha= a[k*stride+i];
      daxpy2(&c[0*stride+i], &b[0*stride+k], dimj, alpha, stride);
    }
  }
#endif 
}


void daxpy(a, b, n, alpha)

double *a;
double *b;
double alpha;
int n;

{
  int i;

  for (i=0; i<n; i++) {
    a[i] += alpha*b[i];
  }
}

void daxpy2(a, b, n, alpha, stride)

double *a;
double *b;
double alpha;
int n;
int stride;

{
  int i;

  for (i=0; i<n; i++) {
    a[i*stride] += alpha*-b[i*stride];
  }
}

int BlockOwner(I, J)

int I;
int J;

{
  return((I%num_cols) + (J%num_rows)*num_cols);
}


void lu(n, bs, MyNum, lc, dostats)

int n;
int bs;
int MyNum;
struct LocalCopies *lc;
int dostats;

{
  int i, il, j, jl, k, kl;
  int I, J, K;
  double *A, *B, *C, *D;
  int dimI, dimJ, dimK;
  int strI;
  unsigned int t1, t2, t3, t4, t11, t22;
  int diagowner;
  int colowner;

  strI = n;
  for (k=0, K=0; k<n; k+=bs, K++) {
    kl = k+bs; 
    if (kl>n) {
      kl = n;
    }

    if ((MyNum == 0) || (dostats)) {
      {long time(long *); (t1) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;
    }

    /* factor diagonal block */
    diagowner = BlockOwner(K, K);
    if (diagowner == MyNum) {
      A = &(a[k+k*n]); 
      lu0(A, kl-k, strI);
    
#ifdef LU_FLAG    
      WRITEFLAG(&flagnum(K,K).flag,1);
#endif    
    }

    if ((MyNum == 0) || (dostats)) {
      {long time(long *); (t11) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;
    }

#ifndef LU_FLAG
    {TREEBAR(&(Global->start), MyNum);};
#endif    

    if ((MyNum == 0) || (dostats)) {
      {long time(long *); (t2) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;
    }

    /* divide column k by diagonal block */
    D = &(a[k+k*n]);
    for (i=kl, I=K+1; i<n; i+=bs, I++) {
      if (BlockOwner(I, K) == MyNum) {  /* parcel out blocks */
        il = i + bs;
        if (il > n) {
          il = n;
        }
        A = &(a[i+k*n]);
#ifdef LU_FLAG
	Kstack = K;
	Istack = I;
#endif
        bdiv(A, D, strI, n, il-i, kl-k);
      }
    }
    /* modify row k by diagonal block */
    for (j=kl, J=K+1; j<n; j+=bs, J++) {
      if (BlockOwner(K, J) == MyNum) {  /* parcel out blocks */
        jl = j+bs;
        if (jl > n) {
          jl = n;
        }
        A = &(a[k+j*n]);
#ifdef LU_FLAG
	Jstack = J;
	Kstack = K;
#endif
        bmodd(D, A, kl-k, jl-j, n, strI);
      }
    }

    if ((MyNum == 0) || (dostats)) {
      {long time(long *); (t22) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;
    }

#ifndef LU_FLAG
    {TREEBAR(&(Global->start), MyNum);};
#endif

    if ((MyNum == 0) || (dostats)) {
      {long time(long *); (t3) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;
    }

    /* modify subsequent block columns */
    for (i=kl, I=K+1; i<n; i+=bs, I++) {
      il = i+bs;
      if (il > n) {
        il = n;
      }
      colowner = BlockOwner(I,K);
      A = &(a[i+k*n]);
      for (j=kl, J=K+1; j<n; j+=bs, J++) {
        jl = j + bs;
        if (jl > n) {
          jl = n;
        }
        if (BlockOwner(I, J) == MyNum) {  /* parcel out blocks */
          B = &(a[k+j*n]);
          C = &(a[i+j*n]);
#ifdef LU_FLAG
	  Istack = I;
	  Jstack = J;
	  Kstack = K;
#endif
          bmod(A, B, C, il-i, jl-j, kl-k, n);
        }
      }
    }
    if ((MyNum == 0) || (dostats)) {
      {long time(long *); (t4) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;
      lc->t_in_fac += (t11-t1);
      lc->t_in_solve += (t22-t2);
      lc->t_in_mod += (t4-t3);
      lc->t_in_bar += (t2-t11) + (t3-t22);
    }
  }
}


void InitA(rhs)

double *rhs;

{
  int i, j;

  srand48((long) 1);
  for (j=0; j<n; j++) {
    for (i=0; i<n; i++) {
      a[i+j*n] = (double) lrand48()/MAXRAND;
      if (i == j) {
	a[i+j*n] *= 10;
      }
    }
  }

  for (j=0; j<n; j++) {
    rhs[j] = 0.0;
  }
  for (j=0; j<n; j++) {
    for (i=0; i<n; i++) {
      rhs[i] += a[i+j*n];
    }
  }
}


double TouchA(bs, MyNum)

int bs;
int MyNum;

{
  int i, j, I, J;
  double tot = 0.0;

  for (J=0; J*bs<n; J++) {
    for (I=0; I*bs<n; I++) {
      if (BlockOwner(I, J) == MyNum) {
        for (j=J*bs; j<(J+1)*bs && j<n; j++) {
          for (i=I*bs; i<(I+1)*bs && i<n; i++) {
            tot += a[i+j*n];
          }
        }
      }
    }
  }
  return(tot);
}


void PrintA()
{
  int i, j;

  for (i=0; i<n; i++) {
    for (j=0; j<n; j++) {
      printf("%8.1f ", a[i+j*n]);
    }
    printf("\n");
  }
}


void CheckResult(n, a, rhs)

int n;
double *a;
double  *rhs;

{
  int i, j, bogus = 0;
  double *y, diff, max_diff;

  y = (double *) malloc(n*sizeof(double));
  if (y == NULL) {
    printerr("Could not malloc memory for y\n");
    exit(-1);
  }
  for (j=0; j<n; j++) {
    y[j] = rhs[j];
  }
  for (j=0; j<n; j++) {
    y[j] = y[j]/a[j+j*n];
    for (i=j+1; i<n; i++) {
      y[i] -= a[i+j*n]*y[j];
    }
  }

  for (j=n-1; j>=0; j--) {
    for (i=0; i<j; i++) {
      y[i] -= a[i+j*n]*y[j];
    }
  }

  max_diff = 0.0;
  for (j=0; j<n; j++) {
    diff = y[j] - 1.0;
    if (fabs(diff) > 0.00001) {
      bogus = 1;
      max_diff = diff;
    }
  }
  if (bogus) {
    printf("TEST FAILED: (%.5f diff)\n", max_diff);
  } else {
    printf("TEST PASSED\n");
  }
  free(y);
}


void printerr(s)

char *s;

{
  fprintf(stderr,"ERROR: %s\n",s);
}

